@@ -15,13 +15,24 @@ limitations under the License.
 #ifndef TENSORFLOW_CORE_DATA_DATASET_UTILS_H_
 #define TENSORFLOW_CORE_DATA_DATASET_UTILS_H_
 
+#include <atomic>
+#include <cstdint>
 #include <functional>
+#include <memory>
 #include <string>
+#include <utility>
+#include <vector>
 
+#include "absl/container/flat_hash_map.h"
 #include "absl/container/flat_hash_set.h"
+#include "absl/status/status.h"
+#include "absl/strings/string_view.h"
 #include "tensorflow/core/common_runtime/function.h"
 #include "tensorflow/core/framework/dataset.h"
+#include "tensorflow/core/framework/dataset_options.pb.h"
 #include "tensorflow/core/framework/function.h"
+#include "tensorflow/core/framework/function.pb.h"
+#include "tensorflow/core/framework/node_def.pb.h"
 #include "tensorflow/core/framework/resource_handle.h"
 #include "tensorflow/core/framework/resource_mgr.h"
 #include "tensorflow/core/framework/tensor.h"
@@ -33,14 +44,12 @@ namespace data {
 // should be supplied by the auto-sharding rewrite.
 constexpr int kShardHint = -1;
 
-// The initial parallelism value before Autotune has a chance to optimize.
-constexpr int kAutotuneDefaultParallelism = 16;
-
 // Creates a resource handle with a unique name for the given resource where
 // the resource is managed by the Resource Manager.
 template <typename T>
-Status CreateWeakHandle(OpKernelContext* ctx, T* resource,
-                        const string& container_name, ResourceHandle* handle) {
+absl::Status CreateWeakHandle(OpKernelContext* ctx, T* resource,
+                              const string& container_name,
+                              ResourceHandle* handle) {
   static std::atomic<int64_t> resource_id_counter(0);
   string unique_name =
       strings::StrCat(container_name, resource_id_counter.fetch_add(1));
@@ -49,19 +58,20 @@ Status CreateWeakHandle(OpKernelContext* ctx, T* resource,
 
   *handle = MakeResourceHandle(container_name, unique_name, *ctx->device(),
                                TypeIndex::Make<T>());
-  return OkStatus();
+  return absl::OkStatus();
 }
 
 // Creates a ref-counting resource handle for the given resource, where the
 // resource is owned by the handle.
 template <typename T>
-Status CreateHandle(OpKernelContext* ctx, T* resource, ResourceHandle* handle) {
+absl::Status CreateHandle(OpKernelContext* ctx, T* resource,
+                          ResourceHandle* handle) {
   ResourceMgr* mgr = ctx->resource_manager();
   *handle =
       ResourceHandle::MakeRefCountingHandle(resource, ctx->device()->name());
   TF_RETURN_IF_ERROR(
       mgr->CreateUnowned<T>(handle->container(), handle->name(), resource));
-  return OkStatus();
+  return absl::OkStatus();
 }
 
 // TODO(b/198162355): Merge this class with ResourceOpKernel.
@@ -118,7 +128,7 @@ class AnonymousResourceOp : public OpKernel {
  protected:
   virtual string name() = 0;
 
-  virtual Status CreateResource(
+  virtual absl::Status CreateResource(
       OpKernelContext* ctx, std::unique_ptr<FunctionLibraryDefinition> flib_def,
       std::unique_ptr<ProcessFunctionLibraryRuntime> pflr,
       FunctionLibraryRuntime* lib, T** resource) = 0;
@@ -128,22 +138,24 @@ class AnonymousResourceOp : public OpKernel {
   const bool return_deleter_;
 };
 
-// Returns Status::OK() if `expected` and `received` types match,
+// Returns OkStatus() if `expected` and `received` types match,
 // errors::InvalidArgument otherwise.
-Status VerifyTypesMatch(const DataTypeVector& expected,
-                        const DataTypeVector& received);
+absl::Status VerifyTypesMatch(const DataTypeVector& expected,
+                              const DataTypeVector& received);
 
-Status VerifyTypesMatch(const DataTypeVector& expected,
-                        const std::vector<Tensor>& received);
-
-// Returns Status::OK() if `expected` and `received` shapes are compatible,
-// errors::InvalidArgument otherwise.
-Status VerifyShapesCompatible(const std::vector<PartialTensorShape>& expected,
-                              const std::vector<PartialTensorShape>& received);
-
-Status VerifyShapesCompatible(const std::vector<PartialTensorShape>& expected,
+absl::Status VerifyTypesMatch(const DataTypeVector& expected,
                               const std::vector<Tensor>& received);
 
+// Returns OkStatus() if `expected` and `received` shapes are compatible,
+// errors::InvalidArgument otherwise.
+absl::Status VerifyShapesCompatible(
+    const std::vector<PartialTensorShape>& expected,
+    const std::vector<PartialTensorShape>& received);
+
+absl::Status VerifyShapesCompatible(
+    const std::vector<PartialTensorShape>& expected,
+    const std::vector<Tensor>& received);
+
 // Dataset op level determinism policy.
 class DeterminismPolicy {
  public:
@@ -167,7 +179,7 @@ class DeterminismPolicy {
   // kNondeterministic, depending on the values of `is_deterministic`.
   explicit DeterminismPolicy(bool is_deterministic);
 
-  static Status FromString(const std::string& s, DeterminismPolicy* out);
+  static absl::Status FromString(const std::string& s, DeterminismPolicy* out);
 
   // Returns the string representing the determinism policy. This will be one of
   // the string constants defined above.
@@ -195,18 +207,18 @@ std::pair<int64_t, int64_t> MaybeOverrideSeeds(
 // Adds the functions in `to_add` to `base`. If a function with a matching
 // signature already exists in `base`, replaces it with the function from
 // `to_add`.
-Status AddToFunctionLibrary(FunctionLibraryDefinition* base,
-                            const FunctionLibraryDefinition& to_add);
-Status AddToFunctionLibrary(FunctionLibraryDefinition* base,
-                            const FunctionDefLibrary& to_add);
+absl::Status AddToFunctionLibrary(FunctionLibraryDefinition* base,
+                                  const FunctionLibraryDefinition& to_add);
+absl::Status AddToFunctionLibrary(FunctionLibraryDefinition* base,
+                                  const FunctionDefLibrary& to_add);
 
 // Determines whether the given function is stateful.
-Status IsFunctionStateful(const FunctionLibraryDefinition& library,
-                          const FunctionDef& function_def);
+absl::Status IsFunctionStateful(const FunctionLibraryDefinition& library,
+                                const FunctionDef& function_def);
 
 // Determines whether the given node is stateful.
-Status IsNodeStateful(const FunctionLibraryDefinition& library,
-                      const NodeDef& node);
+absl::Status IsNodeStateful(const FunctionLibraryDefinition& library,
+                            const NodeDef& node);
 
 // Creates a runner that runs functions with limited parallelism.
 std::function<void(std::function<void()>)> RunnerWithMaxParallelism(
@@ -240,7 +252,8 @@ class DummyResourceOp : public OpKernel {
 // MatchesAnyVersion("BatchDataset", "BatchDatasetV2") == true
 // MatchesAnyVersion("BatchDataset", "BatchDatasetV3") == true
 // MatchesAnyVersion("PaddedBatchDataset", "BatchDataset") == false
-bool MatchesAnyVersion(StringPiece op_prefix, StringPiece op_to_match);
+bool MatchesAnyVersion(absl::string_view op_prefix,
+                       absl::string_view op_to_match);
 
 // Returns the index-th slice of a given tensor. If the index-th slice of
 // the tensor is not aligned, returns a deep copy of the tensor.
@@ -250,73 +263,55 @@ Tensor MaybeCopySubSlice(const Tensor& tensor, int64 index);
 void StripDevicePlacement(FunctionDefLibrary* library);
 
 // Copies partial of the batch output.
-Status CopyPartialBatch(int64_t num_elements, const Tensor& value,
-                        Tensor* output);
+absl::Status CopyPartialBatch(int64_t num_elements, const Tensor& value,
+                              Tensor* output);
 
 // Reads a batch when restoring the iterator.
-Status ReadBatch(IteratorContext* ctx, IteratorStateReader* reader,
-                 int64_t batch_size, const string& iterator_prefix,
-                 const string& batch_prefix, std::vector<Tensor>* batch);
+absl::Status ReadBatch(IteratorContext* ctx, IteratorStateReader* reader,
+                       int64_t batch_size, const string& iterator_prefix,
+                       const string& batch_prefix, std::vector<Tensor>* batch);
 
 // Writes a batch when saving the iterator.
-Status WriteBatch(int64_t batch_size, int64_t num_elements,
-                  const string& iterator_prefix, const string& batch_prefix,
-                  IteratorStateWriter* writer, std::vector<Tensor>* batch);
+absl::Status WriteBatch(int64_t batch_size, int64_t num_elements,
+                        const string& iterator_prefix,
+                        const string& batch_prefix, IteratorStateWriter* writer,
+                        std::vector<Tensor>* batch);
 
 // Reads a status when restoring the iterator.
-Status ReadStatus(const string& iterator_prefix, const string& prefix,
-                  IteratorStateReader* reader, Status* status);
+absl::Status ReadStatus(const string& iterator_prefix, const string& prefix,
+                        IteratorStateReader* reader, absl::Status* status);
 
 // Writes a status when saving the iterator.
-Status WriteStatus(const string& iterator_prefix, const string& prefix,
-                   const Status& status, IteratorStateWriter* writer);
+absl::Status WriteStatus(const string& iterator_prefix, const string& prefix,
+                         const absl::Status& status,
+                         IteratorStateWriter* writer);
 
 // Processes a batch to output. In the case a partial batch is encountered, copy
 // only partial of the batch.
-Status ProcessBatch(int64_t batch_size, int64_t num_elements,
-                    bool drop_remainder, const Status& status,
-                    IteratorContext* ctx, std::vector<Tensor>* output,
-                    bool* end_of_sequence, std::vector<Tensor>* batch);
-
-// Constructs and stores the parameters for the CopyBatch function.
-struct CopyBatchParams {
-  Allocator* allocator;
-  std::function<void(std::function<void()>)>* runner;
-  int64 runner_threadpool_size;
-
-  explicit CopyBatchParams(IteratorContext* ctx) {
-    allocator = ctx->allocator({});
-    runner = ctx->runner();
-    runner_threadpool_size = ctx->runner_threadpool_size();
-  }
-
-  explicit CopyBatchParams(OpKernelContext* ctx) {
-    allocator = ctx->get_allocator({});
-    runner = ctx->runner();
-    runner_threadpool_size = GetRunnerThreadpoolSizeFromOpKernelContext(ctx);
-  }
-};
+absl::Status ProcessBatch(int64_t batch_size, int64_t num_elements,
+                          bool drop_remainder, const absl::Status& status,
+                          IteratorContext* ctx, std::vector<Tensor>* output,
+                          bool* end_of_sequence, std::vector<Tensor>* batch);
 
 // Copies the input elements to a batch.
 //
 // The `batch_elements` argument contains the individual elements to copy into a
 // batch. The `parallel_copy` argument indicates whether to parallelize the
-// copy. The `allocation_callback` argument can be used to pass a callback to
-// invoke upon successful allocation of the memory for the batch. The
-// `out_tensors` argument will be used to store the resulting batch (one for
+// copy.
+// The `out_tensors` argument will be used to store the resulting batch (one for
 // each component of the input).
-Status CopyBatch(CopyBatchParams params,
-                 const std::vector<std::vector<Tensor>>& batch_elements,
-                 bool parallel_copy,
-                 std::function<Status()> allocation_callback,
-                 std::vector<Tensor>* out_tensors);
+absl::Status CopyBatch(AnyContext ctx,
+                       std::vector<std::vector<Tensor>>&& batch_elements,
+                       bool parallel_copy, std::vector<Tensor>* out_tensors);
 
-// Computes the set of experiments to apply based on the job name, rollout
-// percentage of registered experiments, and the TF_DATA_EXPERIMENT_OPT_IN and
-// TF_DATA_EXPERIMENT_OPT_OUT environment variables.
+// Computes the set of experiments to apply based on the job name, task id,
+// rollout percentage of registered experiments, and the
+// TF_DATA_EXPERIMENT_OPT_IN and TF_DATA_EXPERIMENT_OPT_OUT environment
+// variables.
 absl::flat_hash_set<string> GetExperiments();
 absl::flat_hash_set<string> GetExperiments(
-    const string& job_name, std::function<uint64(const string&)> hash_func);
+    const std::string& job_name, int64_t task_id,
+    std::function<uint64_t(const string&)> hash_func);
 
 // Logs and records the experiments that will be applied.
 void LogAndRecordExperiments(const absl::flat_hash_set<string>& experiments);
@@ -361,35 +356,73 @@ inline int GetCpuBudget() {
 // optimization.
 int64 GetAutotuneDefaultParallelism(IteratorContext* ctx);
 
+// Creates an iterator context appropriate for a nested dataset's iterator. A
+// nested dataset is a dataset created within another dataset, e.g. by the
+// function passed to `interleave` or `flat_map`.
+IteratorContext MakeNestedIteratorContext(IteratorContext* ctx);
+
+// A `DatasetExperimentRegistry::JobSelector` that randomly selects
+// `rollout_pct` percent of all jobs. `name_hash` is a hash of the experiment
+// and job names.
+template <int64_t rollout_pct>
+bool RandomJobSamplePercentage(uint64_t name_hash) {
+  return name_hash % 100 < rollout_pct;
+}
+
+// A `DatasetExperimentRegistry::TaskSelector` that selects all tasks.
+bool AllTasks(int64_t unused_task_id, bool unused_evens);
+
+// A `DatasetExperimentRegistry::TaskSelector` that selects the tasks for half
+// of all hosts. Typically, one or two consecutive tasks run on a single host.
+// If `evens` is `true`, selects tasks 0,1,4,5,8,9,..., otherwise selects tasks
+// 2,3,6,7,10,11,...
+bool IndependentHostTasks(int64_t task_id, bool evens);
+
 // Registry of tf.data experiments.
 class DatasetExperimentRegistry {
  public:
+  using JobSelector = std::function<bool(uint64_t name_hash)>;
+  using TaskSelector = std::function<bool(int64_t task_id, bool evens)>;
+
+  struct ExperimentSelector {
+    JobSelector job_selector;
+    TaskSelector task_selector;
+  };
+
   // Registers the experiment.
-  static void Register(const string& experiment, int64_t rollout_pct);
+  static void Register(const string& experiment, JobSelector job_selector,
+                       TaskSelector task_selector);
 
   // Returns all registered experiments.
-  static absl::flat_hash_map<string, int64_t> Experiments();
+  static absl::flat_hash_map<string, ExperimentSelector> Experiments();
 };
 
 // Helper class to register a dataset experiment.
 class DatasetExperimentRegistrar {
  public:
-  explicit DatasetExperimentRegistrar(const string& experiment,
-                                      int64_t rollout_pct) {
-    DatasetExperimentRegistry::Register(experiment, rollout_pct);
+  explicit DatasetExperimentRegistrar(
+      const string& experiment,
+      DatasetExperimentRegistry::JobSelector job_selector,
+      DatasetExperimentRegistry::TaskSelector task_selector) {
+    DatasetExperimentRegistry::Register(experiment, job_selector,
+                                        task_selector);
   }
 };
 
 // Macro that can be used to register a dataset experiment.
-#define REGISTER_DATASET_EXPERIMENT(experiment, rollout_pct) \
-  REGISTER_DATASET_OP_NAME_UNIQ_HELPER(__COUNTER__, experiment, rollout_pct)
+#define REGISTER_DATASET_EXPERIMENT(experiment, job_selector, task_selector)  \
+  REGISTER_DATASET_OP_NAME_UNIQ_HELPER(__COUNTER__, experiment, job_selector, \
+                                       task_selector)
 
-#define REGISTER_DATASET_OP_NAME_UNIQ_HELPER(ctr, experiment, rollout_pct) \
-  REGISTER_DATASET_OP_NAME_UNIQ(ctr, experiment, rollout_pct)
+#define REGISTER_DATASET_OP_NAME_UNIQ_HELPER(ctr, experiment, job_selector, \
+                                             task_selector)                 \
+  REGISTER_DATASET_OP_NAME_UNIQ(ctr, experiment, job_selector, task_selector)
 
-#define REGISTER_DATASET_OP_NAME_UNIQ(ctr, experiment, rollout_pct) \
-  static ::tensorflow::data::DatasetExperimentRegistrar             \
-      registrar__body__##ctr##__object(experiment, rollout_pct)
+#define REGISTER_DATASET_OP_NAME_UNIQ(ctr, experiment, job_selector, \
+                                      task_selector)                 \
+  static ::tensorflow::data::DatasetExperimentRegistrar              \
+      registrar__body__##ctr##__object(experiment, job_selector,     \
+                                       task_selector)
 
 }  // namespace data
 }  // namespace tensorflow
